/* $Id: sfskey.C,v 1.19 1999/04/25 04:44:16 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfskey.h"

typedef enum { MODE_ADD, MODE_LIST, MODE_KILL, MODE_REM_ALL, MODE_REM,
	       MODE_FETCH, MODE_KEYGEN, MODE_SRPGEN, MODE_REGISTER, 
	       MODE_UPDATE, MODE_HOSTID } mode_type;

static void
usage ()
{
  warnx << "usage: " << progname << " [command]\n\n";

  warnx << "\tadd [-S sock] [keyfile | [user]@authservhostname]\n";
  warnx << "\tlist [-S sock]\n";
  warnx << "\tkill [-S sock]\n";
  warnx << "\tremall [-S sock]\n";
  warnx << "\trem [-S sock] [keyfile]\n";
  warnx << "\tfetch [user]@authservhostname keyfile\n";
  warnx << "\tgen [-b bits] [-P] [-k] [keyfile]\n";
  warnx << "\tgensrp [-b bits] srpparamfile\n";
  warnx << "\treg [-s] [-R srpparamfile] [keyfile]\n";
  warnx << "\tupdate [-s] [-R srpparamfile] [-a authserv] " 
	<< "[oldkeyfile [newkeyfile]]\n";
  warnx << "\thostid [-a] [sfsserverhostname]\n";
  exit (1);
}

static void
boolexit (bool b)
{
  exit (!b);
}

static void
do_hostid (str hostname, str path)
{
  if (!path) {
    warn << "Unable to fetch hostid from server.\n";
    exit (1);
  }

  if (hostname)
    warnx << "WARNING!!!  THE FOLLOWING HOSTID WAS RETRIEVED INSECURELY "
	  << "OVER THE NETWORK.\nUSE IT AT YOUR OWN RISK.\n";

  warnx << path << "\n";
  exit (0);
}

static void
parseopt_agent (int argc, char **argv, str *keyfile)
{
  int ch;
  while ((ch = getopt (argc, argv, "S:")) != -1)
    switch (ch) {
    case 'S':
      if (agentsock)
	usage ();
      agentsock = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 1) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 1)
    *keyfile = argv[0];
  else
    *keyfile = dotsfs << "/identity";
}

static void
parseopt_fetch (int argc, char **argv, str *keyfile, 
		str *authserv, str *username)
{
  *authserv = NULL;

  int ch;
  while ((ch = getopt (argc, argv, "")) != -1)
    switch (ch) {
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 2) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 2) {
    *authserv = argv[0];
    *keyfile = argv[1];
  }
  else {
    warn << "too few arguments\n";
    usage ();
  }

  if (!parse_userhost (*authserv, username, authserv))
    fatal << "Specify key locate as [user]@authservhostname please.\n";
}

static void
parseopt_keygen (int argc, char **argv, str *keyfile, str *comment, 
		 int *bits, bool *opt_nopwd, bool *opt_kbdnoise, bool *opt_b)
{
  enum { minsize = 768 };

  *bits = 1024;
  *comment = NULL;
  *opt_nopwd = false;
  *opt_kbdnoise = false;
  *opt_b = false;

  int ch;
  while ((ch = getopt (argc, argv, "Pkb:c:")) != -1)
    switch (ch) {
    case 'b':
      *opt_b = true;
      *bits = atoi (optarg);
      if (*bits < minsize)
	fatal ("bits specified with -b must be at least %d\n", minsize);
      break;
    case 'P':
      *opt_nopwd = true;
      break;
    case 'k':
      *opt_kbdnoise = true;
      break;
    case 'c':
      *comment = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 1) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 1)
    *keyfile = argv[0];
  else
    *keyfile = dotsfs << "/identity";
}

static void
parseopt_srpgen (int argc, char **argv, str *srpfile, int *bits)
{
  enum { minsize = 768 };

  *srpfile = NULL;
  *bits = 1024;

  int ch;
  while ((ch = getopt (argc, argv, "b:c:")) != -1)
    switch (ch) {
    case 'b':
      *bits = atoi (optarg);
      if (*bits < minsize)
	fatal ("bits specified with -b must be at least %d\n", minsize);
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc != 1) {
    warn << "wrong number of arguments\n";
    usage ();
  }
  else
    *srpfile = argv[0];
}

static void
parseopt_register (int argc, char **argv, str *keyfile, 
		   str *srpfile, bool *opt_sec)
{
  *opt_sec = false;
  *srpfile = NULL;

  int ch;
  while ((ch = getopt (argc, argv, "sR:")) != -1)
    switch (ch) {
    case 's':
      *opt_sec = true;
      break;
    case 'R':
      *opt_sec = true;
      *srpfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 1) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 1)
    *keyfile = argv[0];
  else
    *keyfile = dotsfs << "/identity";

  if (*srpfile)
    warnx << "SRP Parameter File: " << *srpfile << "\n";
}

static void
parseopt_update (int argc, char **argv, str *oldkf, str *newkf, 
		 str *srpfile, bool *opt_sec, str *authserv)
{
  *opt_sec = false;
  *authserv = NULL;
  *srpfile = NULL;

  int ch;
  while ((ch = getopt (argc, argv, "sa:R:")) != -1)
    switch (ch) {
    case 's':
      *opt_sec = true;
      break;
    case 'a':
      *authserv = optarg;
      break;
    case 'R':
      *opt_sec = true;
      *srpfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 2) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 2) {
    *oldkf = argv[0];
    *newkf = argv[1];
  }
  else if (argc == 1) {
    *oldkf = argv[0];
    *newkf = dotsfs << "/identity";
  }
  else {
    *oldkf = dotsfs << "/identity";
    *newkf = dotsfs << "/identity";
  }

  if (!*authserv && !(*authserv = file2str (dotsfs << "/authserv")))
    fatal << "No authserv specified in " << dotsfs << "/authserv\n"
	  << "or with -a command line option.\n";

  rxx r (rxx ("^([A-Za-z0-9\\.\\-]+:.+)$"));
  if (r.match (*authserv)) {
    *authserv = r[1];
    warnx << "Authentication Server: " << *authserv << "\n";
  }
  else {
    if (*authserv == "-") {
      warnx << "Authentication Server: local via Unix domain socket.\n";
      *authserv = NULL;
    }
    else
      fatal << "Incorrectly formated authserv:\n" 
	    << "\tUse hostname:hostid format to contact a remote server.\n"
	    << "\tUse '-' to contact the local authserv through a Unix\n"
	    << "\tdomain socket.\n";
  }

  if (*srpfile)
    warnx << "SRP Parameter File: " << *srpfile << "\n";
}

static void
parseopt_hostid (int argc, char **argv, str *hostname, bool *opt_a)
{
  *hostname = NULL;
  *opt_a = false;

  int ch;
  while ((ch = getopt (argc, argv, "a")) != -1)
    switch (ch) {
    case 'a':
      *opt_a = true;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 1) {
    warn << "too many arguments\n";
    usage ();
  }
  else if (argc == 1) {
    *hostname = argv[0];
  }
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);
  mode_type mode = MODE_ADD;

  if (argc < 2)
    usage ();
  else if (!strcmp (argv[1], "add"))
    mode = MODE_ADD;
  else if (!strcmp (argv[1], "list"))
    mode = MODE_LIST;
  else if (!strcmp (argv[1], "kill"))
    mode = MODE_KILL;
  else if (!strcmp (argv[1], "remall"))
    mode = MODE_REM_ALL;
  else if (!strcmp (argv[1], "rem"))
    mode = MODE_REM;
  else if (!strcmp (argv[1], "fetch"))
    mode = MODE_FETCH;
  else if (!strcmp (argv[1], "gen"))
    mode = MODE_KEYGEN;
  else if (!strcmp (argv[1], "gensrp"))
    mode = MODE_SRPGEN;
  else if (!strcmp (argv[1], "reg"))
    mode = MODE_REGISTER;
  else if (!strcmp (argv[1], "update"))
    mode = MODE_UPDATE;
  else if (!strcmp (argv[1], "hostid"))
    mode = MODE_HOSTID;
  else
    usage ();

  argc -= 1;
  argv += 1;

  usrdir_set ();
  random_init ();

  switch (mode) {
  case MODE_ADD:
  case MODE_LIST:
  case MODE_KILL:
  case MODE_REM_ALL:
  case MODE_REM:
    {
      str keyfile, username, authserv;
      parseopt_agent (argc, argv, &keyfile);

      if (mode != MODE_KILL)
	agentc_init ();

      if (parse_userhost (keyfile, &username, &authserv)) {
	vNew keyfetch (NULL, authserv, username, wrap (exit));
	break;
      }

      switch (mode) {
      case MODE_ADD:
	addkey (agentc, keyfile, wrap (boolexit));
	break;
      case MODE_LIST:
	dolist (agentc, wrap (boolexit));
	break;
      case MODE_KILL:
	kill_agent (wrap (boolexit));
	break;
      case MODE_REM:
	removekey (agentc, keyfile, wrap (boolexit));
	break;
      case MODE_REM_ALL:
	removeallkeys (agentc, wrap (boolexit));
	break;
      default:
	fatal << "unimplemented\n";
      }

      break;
    }
  case MODE_FETCH:
    {
      str keyfile, authserv, username;

      parseopt_fetch (argc, argv, &keyfile, &authserv, &username);
      vNew keyfetch (keyfile, authserv, username, wrap (exit));
      break;
    }
  case MODE_KEYGEN:
    {
      str keyfile, comment;
      int bits;
      bool opt_nopwd, opt_kbdnoise, opt_b;

      parseopt_keygen (argc, argv, &keyfile, &comment,
		       &bits, &opt_nopwd, &opt_kbdnoise, &opt_b);
      vNew keygen (keyfile, comment, bits, opt_nopwd, 
		   opt_kbdnoise, opt_b, wrap (exit));
      break;
    }
  case MODE_SRPGEN:
    {
      str srpfile;
      int bits;
      srp s;

      parseopt_srpgen (argc, argv, &srpfile, &bits);

      warnx << "Generating SRP Parameter File: " << srpfile << "\n";
      warnx << "This may take a while...\n";
      s.gen_params (bits);

      str2file (srpfile, export_srp_params (s), 0644);
      exit (0);
    }
  case MODE_REGISTER:
    {
      str keyfile, srpfile;
      bool opt_sec;

      parseopt_register (argc, argv, &keyfile, &srpfile, &opt_sec);
      vNew keyregister (keyfile, srpfile, opt_sec, wrap (exit));
      break;
    }
  case MODE_UPDATE:
    {
      str oldkeyfile, newkeyfile, srpfile, authserv;
      bool opt_sec;

      parseopt_update (argc, argv, &oldkeyfile, &newkeyfile, 
		       &srpfile, &opt_sec, &authserv);
      vNew keyupdate (oldkeyfile, newkeyfile, srpfile, opt_sec, authserv,
		      wrap (exit));
      break;
    }
  case MODE_HOSTID:
    {
      str hostname;
      bool opt_a;
      sfs_service service = SFS_SFS;

      parseopt_hostid (argc, argv, &hostname, &opt_a);
      if (opt_a)
	service = SFS_AUTHSERV;

      sfs_gethostid (hostname, service, wrap (do_hostid, hostname));
    }
  }
  amain ();
}
