// -*-c++-*-
/* $Id: afsnode.h,v 1.17 1999/03/25 06:05:42 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "arpc.h"
#include "nfs_prot.h"
#include "nfs3_prot.h"
#include "itree.h"
#include "vector.h"
#include "str.h"
#include "sfsmisc.h"

#define fattr3 fattr3exp

/* The preferable "ino_t" is already in use on some systems */
typedef u_int64_t inum_t;
extern char fhsecret[16];

void getnfstime (nfstime3 *ntp);

inline bool
operator< (const nfstime3 &a, const nfstime3 &b)
{
  if (a.seconds < b.seconds)
    return true;
  if (a.seconds == b.seconds)
    return a.nseconds < b.nseconds;
  return false;
}

extern void nfs3_err (svccb *sbp, nfsstat3 status);
inline void
nfs_error (svccb *sbp, const nfsstat err)
{
  if (sbp->vers () != 3) {
    sbp->reply (&err);
    return;
  }
  nfsstat3 stat = nfsstat3 (err);
  if (stat == EINVAL)
    stat = NFS3ERR_INVAL;
  nfs3_err (sbp, stat);
}

str fh2str (const nfs_fh *);
str au2str (const authunix_parms *);

class afsnode : public virtual refcount {
protected:
  u_int nlinks;
  nfstime3 mtime;
  nfstime3 ctime;

  afsnode (ftype3, inum_t = 0);
  void lookup_reply (svccb *sbp, afsnode *e);

public:
  ihash_entry<afsnode> itblink;

  const ftype3 type;
  const inum_t ino;

  virtual ~afsnode ();

  virtual u_int getnlinks () { return nlinks; }
  virtual void bumpctime () { getnfstime (&ctime); }
  void addlink () { nlinks++; bumpctime (); }
  void remlink () { nlinks--; bumpctime (); }

  const nfstime3 &getmtime () const { return mtime; }
  const nfstime3 &getctime () const { return ctime; }

  virtual void mkfh (nfs_fh *);
  virtual bool chkfh (const nfs_fh *);
  static afsnode *fh2node (const nfs_fh *);

  void mkfh3 (nfs_fh3 *);
  bool chkfh3 (const nfs_fh *);
  static afsnode *fh3node (const nfs_fh3 *);

  virtual void mkfattr (fattr *, const svccb *);
  virtual void mkfattr3 (fattr3 *, const svccb *);
  void mkpoattr (post_op_attr &poa, const svccb *sbp)
    { poa.set_present (true); mkfattr3 (poa.attributes.addr (), sbp); }

  virtual void nfs_getattr (svccb *);
  virtual void nfs_lookup (svccb *, str);
  virtual void nfs_readlink (svccb *);
  virtual void nfs_read (svccb *);
  virtual void nfs_readdir (svccb *);
  virtual void nfs_statfs (svccb *);
  static void dispatch (svccb *);

  virtual void nfs3_access (svccb *);
  virtual void nfs3_fsstat (svccb *);
  virtual void nfs3_fsinfo (svccb *);
  static void dispatch3 (svccb *);
};

class afslink : public afsnode {
  vector<svccb *> sbps;

  bool resok;
  readlinkres res;
  sfs_aid lastaid;

  void reply ();

protected:
  afslink ();
  virtual ~afslink ();
  void sendreply (svccb *sbp);

public:
  void bumpctime () { ctime.seconds++; }
  void mkfattr3 (fattr3 *, const svccb *);
  void setres (nfsstat err);
  void setres (nfspath path);

  void nfs_readlink (svccb *sbp);

  static ref<afslink> alloc () { return New refcounted<afslink>; }
  static ref<afslink> alloc (nfspath res) {
    ref<afslink> ret = alloc ();
    ret->setres (res);
    return ret;
  }
};

class mntfs;
class delaypt : public afsnode {
  const str name;
  ref<afslink> wlink;
  ptr<mntfs> mdir;
  ptr<afslink> mlink;

  bool resok;
  bool finalized;
  str delaypath;
  sfs_aid lastaid;

  void delthis () { delete this; }
  bool trydel ();
  void getmntfs (ref<mntfs>);

protected:
  delaypt ();
  PRIVDEST ~delaypt ();
  void bumpctime () { ctime.seconds++; }
  void mkfattr3 (fattr3 *, const svccb *);
  void finalize ();

public:
  static ref<delaypt> alloc () { return New refcounted<delaypt>; }
  void nfs_getattr (svccb *sbp);
  void nfs_readlink (svccb *sbp);
  void nfs3_readlink (svccb *);
  void setres (str);
  void setres (nfsstat err);
};

class afsdir;
struct afsdirentry {
  afsdir *const dir;
  const str name;
  const ref<afsnode> node;
  const u_int32_t cookie;

  ihash_entry<afsdirentry> clink;
  itree_entry<afsdirentry> dlink;

  afsdirentry (afsdir *dir, const str &name, afsnode *node);
  ~afsdirentry ();

  static void del (afsdirentry *e) { delete e; }
};

class afsdir : public afsnode {
  friend class afsdirentry;

  itree<const str, afsdirentry, &afsdirentry::name,
    &afsdirentry::dlink> entries;
  afsdir *const parent;

protected:
  afsdir (afsdir *parent);
  virtual ~afsdir ();

  static BOOL xdr (XDR *, void *);

public:
  virtual void bumpmtime () { getnfstime (&mtime); }
  virtual afsnode *lookup (const str &name, const svccb *sbp);

  virtual bool entryok (afsdirentry *, const svccb *);
  virtual afsdirentry *firstentry (const svccb *);
  virtual afsdirentry *nextentry (afsdirentry *, const svccb *);

  virtual bool link (afsnode *node, const str &name);
  virtual bool unlink (const str &name);
  virtual ptr<afsdir> mkdir (const str &name);
  virtual ptr<afslink> symlink (const str &contents, const str &name);

  virtual void nfs_lookup (svccb *, str);
  virtual void nfs_readdir (svccb *sbp)
    { sbp->reply (sbp, &xdr); }

  // XXX - extremely fragile.  If New refcounted<afsdir> appears more
  // than once in any translation unit it causes an internal compiler
  // error in gcc.  Be sure to test with gcc when changing this class.
  static ptr<afsdir> allocsubdir (afsdir *parent)
    { return New refcounted<afsdir> (parent); }
  static ptr<afsdir> alloc () { return allocsubdir (NULL); }

  void ls ();
};
