/* $Id: client.C,v 1.41 1999/05/17 06:54:50 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsrwsd.h"
#include <grp.h>

ihash<const u_int64_t, client, &client::generation, &client::glink> clienttab;

void
client::fail ()
{
  sfsx = NULL;
  nfssrv = NULL;
  sfssrv = NULL;
}

void
client::nfs3reply (svccb *sbp, void *res, filesrv::reqstate rqs, clnt_stat err)
{
  auto_xdr_delete axd (nfs_program_3.tbl[sbp->proc ()].xdr_res, res);

  if (err) {
    sbp->reject (SYSTEM_ERR);
    return;
  }
  if (fsrv->fixres (sbp, res, &rqs)) {
    doleases (this, sbp, res);
    sbp->reply (res);
  }
}

void
client::nfs3dispatch (svccb *sbp)
{
  if (!sbp) {
    fail ();
    return;
  }
  if (sbp->proc () == NFSPROC3_NULL) {
    sbp->reply (NULL);
    return;
  }

  u_int32_t authno = sbp->getaui ();
  if (authno >= authtab.size () || !authtab[authno]) {
    sbp->reject (AUTH_REJECTEDCRED);
    return;
  }
  if (!fsrv) {
    nfs3exp_err (sbp, NFS3ERR_BADHANDLE);
    return;
  }

  filesrv::reqstate rqs;
  if (!fsrv->fixarg (sbp, &rqs))
    return;

  void *res = nfs_program_3.tbl[sbp->proc ()].alloc_res ();
  fsrv->c->call (sbp->proc (), sbp->template getarg<void> (), res,
		 wrap (mkref (this), &client::nfs3reply, sbp, res, rqs),
		 authtab[authno]);
}

void
client::dologin (svccb *sbp)
{
  if (!authid_valid || !fsrv->getauthclnt ()) {
    sbp->replyref (sfs_loginres (SFSLOGIN_ALLBAD));
    return;
  }
  sfsauth_loginres *resp = New sfsauth_loginres;
  fsrv->authclnt->call (SFSAUTHPROC_LOGIN,
			sbp->template getarg<sfs_loginarg> (), resp,
			wrap (mkref (this), &client::dologin_cb, sbp, resp));
}

void
client::dologin_cb (svccb *sbp, sfsauth_loginres *_resp, clnt_stat stat)
{
  auto_ptr<sfsauth_loginres> resp (_resp);
  if (stat) {
    warn << "authserv: " << stat << "\n";
    sbp->replyref (sfs_loginres (SFSLOGIN_ALLBAD));
    return;
  }

  sfs_loginres res (resp->status);
  switch (resp->status) {
  case SFSLOGIN_OK:
    {
      if (resp->resok->authid != authid
	  || !seqstate.check (resp->resok->seqno)
	  || resp->resok->cred.type != SFS_UNIXCRED) {
	res.set_status (SFSLOGIN_BAD);
	break;
      }

      size_t authno = authalloc ();
      *res.authno = authno;
      sfs_unixcred &uc = *resp->resok->cred.unixcred;
      authtab[authno] = authunixint_create ("localhost", uc.uid, uc.gid,
					    uc.groups.size (),
					    uc.groups.base ());
      credtab[authno] = resp->resok->cred;
      break;
    }
  case SFSLOGIN_MORE:
    *res.resmore = *resp->resmore;
    break;
  default:
    break;
  }
  sbp->reply (&res);
}

// XXX - MAJOR DEADLOCK PROBLEMS HERE
// XXX - should never call getpw* or getgr*
void
client::doidnames (svccb *sbp)
{
  u_int32_t authno = sbp->getaui ();
  if (!authno || authno >= authtab.size () || !authtab[authno]) {
    sbp->reject (AUTH_REJECTEDCRED);
    return;
  }

  sfs_idnums *argp = sbp->template getarg<sfs_idnums> ();
  sfs_idnames res;
  if (argp->uid != -1)
    if (struct passwd *p = getpwuid (argp->uid)) {
      res.uidname.set_present (true);
      *res.uidname.name = p->pw_name;
    }
  if (argp->gid != -1)
    if (struct group *g = getgrgid (argp->gid)) {
      res.gidname.set_present (true);
      *res.gidname.name = g->gr_name;
    }
  sbp->reply (&res);
}

// XXX - MAJOR DEADLOCK PROBLEMS HERE
// XXX - should never call getpw* or getgr*
void
client::doidnums (svccb *sbp)
{
  u_int32_t authno = sbp->getaui ();
  if (!authno || authno >= authtab.size () || !authtab[authno]) {
    sbp->reject (AUTH_REJECTEDCRED);
    return;
  }

  sfs_idnames *argp = sbp->template getarg<sfs_idnames> ();
  sfs_idnums res = { -1, -1 };
  if (argp->uidname.present)
    if (struct passwd *p = getpwnam (argp->uidname.name->cstr ()))
      res.uid = p->pw_uid;
  if (argp->gidname.present)
    if (struct group *g = getgrnam (argp->gidname.name->cstr ()))
      res.gid = g->gr_gid;
  sbp->reply (&res);
}

void
client::sfsdispatch (svccb *sbp)
{
  if (!sbp) {
    fail ();
    return;
  }

  switch (sbp->proc ()) {
  case SFSPROC_NULL:
    sbp->reply (NULL);
    return;
  case SFSPROC_CONNECT:
    {
      fsrv = defsrv;
      sfs_connectres res (SFS_OK);
      res.reply->release = SFS_RELEASE;
      res.reply->host = fsrv->hinfo;
      res.reply->prog = ex_NFS_PROGRAM;
      res.reply->vers = ex_NFS_V3;
      bzero (&res.reply->charge, sizeof (sfs_hashcharge));
      sbp->reply (&res);
      return;
    }
  }

  /* Can't do anything else before a connect */
  if (!fsrv) {
    sbp->reject (PROC_UNAVAIL);
    return;
  }

  switch (sbp->proc ()) {
  case SFSPROC_GETFSINFO:
    sbp->replyref (fsrv->fsinfo);
    break;
  case SFSPROC_ENCRYPT:
    sfs_server_crypt (sbp, fsrv->sk, &sessid);
    sfs_get_authid (&authid, SFS_SFS, fsrv->hinfo.hostname,
		    &fsrv->hostid, &sessid);
    authid_valid = true;
    break;
  case SFSPROC_LOGIN:
    dologin (sbp);
    break;
  case SFSPROC_LOGOUT:
    authfree (*sbp->template getarg<u_int32_t> ());
    sbp->reply (NULL);
    break;
  case SFSPROC_IDNAMES:
    doidnames (sbp);
    break;
  case SFSPROC_IDNUMS:
    doidnums (sbp);
    break;
  case SFSPROC_GETCRED:
    {
      u_int32_t authno = sbp->getaui ();
      if (authno < credtab.size ())
	sbp->replyref (credtab[authno]);
      else
	sbp->replyref (sfs_cred (SFS_NOCRED));
      break;
    }
  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

size_t
client::authalloc ()
{
  if (authfreelist.size ())
    return authfreelist.pop_back ();
  authtab.push_back ();
  credtab.push_back ();
  return authtab.size () - 1;
}
void
client::authfree (size_t n)
{
  if (n && n < authtab.size () && authtab[n]) {
    authtab[n] = NULL;
    credtab[n].set_type (SFS_NOCRED);
    authfreelist.push_back (n);
  }
}

u_int64_t
client::nextgen ()
{
  static u_int64_t g;
  while (clienttab[++g] || !g)
    ;
  return g;
}

client::client (ref<axprt_crypt> x)
  : fsrv (NULL), sfsx (x), seqstate (128),
    authid_valid (false), generation (nextgen ())
{
  nfssrv = asrv::alloc (sfsx, ex_nfs_program_3,
			wrap (mkref (this), &client::nfs3dispatch));
  sfssrv = asrv::alloc (sfsx, sfs_program_1,
			wrap (mkref (this), &client::sfsdispatch));
  nfscbc = aclnt::alloc (sfsx, ex_nfscb_program_3);
  authtab.push_back () = authunix_create ("localhost", (uid_t) -1,
					  (gid_t) -1, 0, NULL);
  credtab.push_back ();
  clienttab.insert (this);
}

client::~client ()
{
  clienttab.remove (this);
}

ptr<axprt_stream>
client_accept (int fd)
{
  if (fd < 0)
    fatal ("EOF from sfssd\n");
  tcp_nodelay (fd);
  ref<axprt_crypt> x = axprt_crypt::alloc (fd);
  client::launch (x);
  return x;
}

void
client_accept_standalone ()
{
  sockaddr_in sin;
  bzero (&sin, sizeof (sin));
  socklen_t sinlen = sizeof (sin);
  int fd = accept (sfssfd, reinterpret_cast<sockaddr *> (&sin), &sinlen);
  if (fd >= 0)
    client_accept (fd);
  else if (errno != EAGAIN)
    fatal ("accept: %m\n");
}
