/* $Id: sfssd.C,v 1.15 1999/03/29 21:50:51 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfssd.h"
#include "parseopt.h"
#include "rxx.h"

str configfile;
list<sfssrv, &sfssrv::link> srvlist;

struct servicemap {
  const char *name;
  int number;
};

servicemap servicetab[] = {
  { "sfs", SFS_SFS },
  { "authserv", SFS_AUTHSERV },
  { NULL, 0 }
};

void
sfssrv::setprivs ()
{
  if ((uid || gid) && setgroups (0, NULL))
    fatal ("could not void grouplist: %m\n");
  if (gid && setgid (*gid))
    warn ("could not setgid (%d): %m\n", *gid);
  if (uid && setgid (*uid))
    warn ("could not setuid (%d): %m\n", *uid);
}

void
sfssrv::getpkt (const char *, ssize_t, const sockaddr *)
{
  /* We shouldn't receive anything, so assume this is an EOF */
  warn << "EOF from " << argv[0] << "\n";
  x = NULL;
}

void
sfssrv::launch ()
{
  x = axprt_unix_aspawnv (argv[0], argv, 0, wrap (this, &sfssrv::setprivs));
  if (x)
    x->setrcb (wrap (this, &sfssrv::getpkt));
}

bool
sfssrv::insert ()
{
  for (sfssrv *ss = srvlist.first; ss; ss = srvlist.next (ss)) {
    if (host != ss->host || service != ss->service)
      continue;
    if (!hostid) {
      if (ss->hostid)
	continue;
    }
    else if (!ss->hostid || memcmp (hostid->base (),
				    ss->hostid->base (), hostid->size ()))
      continue;
    if (release_min > ss->release_max || release_max < ss->release_min)
      continue;
    warnx << configfile << ":" << line << ": declaration overlaps "
	  << ss->argv[0] << " on line " << ss->line << "\n";
    return false;
  }
  srvlist.insert_head (this);
  return true;
}

static rxx versrx ("^(\\d+)(-(\\d+))?$");
static void
parseconfig ()
{
  str cf = configfile;
  parseargs pa (cf);
  bool errors = false;

  str hostname;
  rpc_ptr<sfs_hash> hostid;
  u_int relmin = 0, relmax = 0;

  int line;
  vector<str> av;
  while (pa.getline (&av, &line)) {
    if (!strcasecmp (av[0], "server")) {
      hostname = myname ();
      hostid.clear ();
      relmin = 0;
      relmax = 0xffffffff;

      str *sp = av.base () + 1, *se = av.lim ();
      if (sp >= se)
	continue;
      if (*sp == "-v") {
	if (++sp >= se) {
	  warn << cf << ":" << line << ": missing argument for -v\n";
	  errors = true;
	  continue;
	}
	if (!versrx.match (*sp++)) {
	  warn << cf << ":" << line << ": bad version range '"
	       << sp[-1] << "'\n";
	  errors = true;
	  continue;
	}
	relmin = atoi (versrx[1]);
	relmax = versrx[3] ? atoi (versrx[3]) : relmin;
      }
      if (sp >= se)
	continue;
      hostname = *sp++;
      if (sp >= se)
	continue;
      hostid.alloc ();
      if (!sfs_ascii2hostid (hostid, *sp)) {
	warn << cf << ":" << line << ": bad host id'" << *sp << "'\n";
	errors = true;
	continue;
      }
      if (++sp < se) {
	warn << cf << ":" << line << ": unexpected argument '" << *sp << "'\n";
	errors = true;
	continue;
      }
    }
    if (!strcasecmp (av[0], "service")) {
      if (av.size () < 3) {
	warn << cf << ":" << line << ": missing arguments to 'Service'\n";
	errors = true;
	continue;
      }
      servicemap *sp = servicetab;
      while (sp->name && strcasecmp (av[1], sp->name))
	sp++;
      if (!sp->name) {
	warn << cf << ":" << line << ": invalid service '" << av[1] << "'\n";
	errors = true;
	continue;
      }

      sfssrv *ss = New sfssrv;
      ss->line = line;
      ss->host = hostname;
      ss->hostid = hostid;
      ss->release_min = relmin;
      ss->release_max = relmax;
      ss->service = sp->number;
      for (str *sp = av.base () + 2; sp < av.lim (); sp++)
	ss->argv.push_back (*sp);
      if (!ss->insert ())
	errors = true;
    }
    else {
      errors = true;
      warn << cf << ":" << line << ": unknown directive '"
	   << av[0] << "'\n";
    }
  }

  if (errors)
    fatal ("parse errors in configuration file\n");
}

static void
clone (ref<asrv> s, ref<axprt_clone> x, svccb *sbp)
{
  s->setcb (NULL);
  if (!sbp)
    return;
  if (sbp->proc () != SFSPROC_CONNECT) {
    sbp->reject (PROC_UNAVAIL);
    return;
  }
  sfs_connectarg *arg = sbp->template getarg<sfs_connectarg> ();

  sfssrv *besthost = NULL;
  sfssrv *bestservice = NULL;
  enum { BAD, HOST, HOSTID, EXACTHOSTID } howgood = BAD;

  for (sfssrv *ss = srvlist.first; ss; ss = srvlist.next (ss)) {
    if (howgood <= HOST && arg->name == ss->host) {
      howgood = HOST;
      besthost = ss;
      if (arg->service == ss->service)
	bestservice = ss;
    }
    if (howgood <= HOSTID && !ss->hostid) {
      howgood = HOSTID;
      besthost = ss;
      if (arg->service == ss->service)
	bestservice = ss;
    }
    if (howgood <= EXACTHOSTID && ss->hostid
	&& !memcmp (arg->hostid.base (), ss->hostid->base (),
		    arg->hostid.size ())) {
      howgood = EXACTHOSTID;
      besthost = ss;
      if (arg->service == ss->service)
	bestservice = ss;
    }
  }

  if (bestservice)
    besthost = bestservice;

  if (besthost) {
    if (besthost->x) {
      besthost->x->clone (x);
      return;
    }
    sbp->replyref (sfs_connectres (SFS_TEMPERR));
    return;
  }
  sbp->replyref (sfs_connectres (SFS_NOSUCHHOST));
}

static void
newserv (int fd)
{
  sockaddr_in sin;
  bzero (&sin, sizeof (sin));
  socklen_t sinlen = sizeof (sin);
  int nfd = accept (fd, (sockaddr *) &sin, &sinlen);
  if (nfd >= 0) {
    warn ("accepted connection from %s\n", inet_ntoa (sin.sin_addr));
    close_on_exec (nfd);
    ref<axprt_clone> x = axprt_clone::alloc (nfd);
    ref<asrv> s = asrv::alloc (x, sfs_program_1);
    s->setcb (wrap (clone, s, x));
  }
  else if (errno != EAGAIN)
    warn ("accept: %m\n");
}

static void
launchservers ()
{
  int fd = inetsocket (SOCK_STREAM, sfs_port);
  if (fd < 0)
    fatal ("could not bind TCP port %d: %m\n", sfs_port);
  close_on_exec (fd);
  listen (fd, 5);
  fdcb (fd, selread, wrap (newserv, fd));

  for (sfssrv *ss = srvlist.first; ss; ss = srvlist.next (ss))
    ss->launch ();
}

static void
usage ()
{
  warnx << progname << ": [-f configfile]\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "xf:")) != -1)
    switch (ch) {
    case 'f':
      if (configfile)
	usage ();
      configfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;
  if (argc > 1)
    usage ();

  sfsconst_init ();
  if (!configfile)
    configfile = sfsconst_etcfile_required ("sfssd_config");

  parseconfig ();
  launchservers ();

  amain ();
}
